<?php

declare(strict_types=1);

namespace FiloBlu\Checkproducts\Model;

use FiloBlu\Checkproducts\Helper\Data;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\Product\Attribute\Source\Status;
use Magento\Catalog\Model\ResourceModel\Product\Collection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Eav\Model\ResourceModel\Entity\Attribute;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\DeadlockException;
use Magento\Framework\EntityManager\MetadataPool;
use Magento\Framework\Exception\LocalizedException;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Model\StoreManagerInterface;

use function count;

/**
 *
 */
class ProductChecker implements ProductCheckerInterface
{
    /** @var int Max retires when deadlock is detected */
    const MAX_RETRIES = 5;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var Data
     */
    protected $helper;

    /**
     * @var CollectionFactory
     */
    private $productCollectionFactory;

    /**
     * @var StoreManagerInterface
     */
    private $storeManager;

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;
    /**
     * @var Attribute
     */
    private $eavAttribute;
    /**
     * @var MetadataPool
     */
    private $metadataPool;


    /**
     * @param CollectionFactory $productCollectionFactory
     * @param Data $helper
     * @param StoreManagerInterface $storeManager
     * @param ScopeConfigInterface $scopeConfig
     * @param ResourceConnection $resourceConnection
     * @param Attribute $eavAttribute
     * @param MetadataPool $metadataPool
     */
    public function __construct(
        CollectionFactory $productCollectionFactory,
        Data $helper,
        StoreManagerInterface $storeManager,
        ScopeConfigInterface $scopeConfig,
        ResourceConnection $resourceConnection,
        Attribute $eavAttribute,
        MetadataPool $metadataPool
    ) {
        $this->scopeConfig = $scopeConfig;
        $this->helper = $helper;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->storeManager = $storeManager;
        $this->resourceConnection = $resourceConnection;
        $this->eavAttribute = $eavAttribute;
        $this->metadataPool = $metadataPool;
    }

    /**
     * @param CollectorInterface $collector
     * @return CollectorInterface
     * @throws LocalizedException
     */
    public function check(CollectorInterface $collector): CollectorInterface
    {
        if (!$this->helper->isEnabled()) {
            return $collector;
        }

        foreach ($this->storeManager->getStores(true) as $store) {
            $this->checkProducts($store, $collector);
        }

        return $collector;
    }

    /**
     * @param StoreInterface $store
     * @param CollectorInterface $collector
     * @return void
     * @throws LocalizedException
     * @throws \Exception
     */
    public function checkProducts(StoreInterface $store, CollectorInterface $collector)
    {
        $metadata = $this->metadataPool->getMetadata(ProductInterface::class);
        $linkField = $metadata->getLinkField();

        $pageSize = $this->helper->getBulkSize();
        $pageNumber = 1;

        do {
            $collection = $this->createCollection($store, $pageNumber, $pageSize);

            if ($collection->count() > 0) {
                $this->processProducts($collection, $linkField, $store, $collector);
            }

            $pageNumber++;
        } while ($pageNumber <= $pageSize);

        $collection->clear();
        unset($collection);
    }

    /**
     * @param StoreInterface $store
     * @param $page
     * @param $pageSize
     * @return Collection
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function createCollection(StoreInterface $store, $page, $pageSize)
    {
        /** @var Collection $productCollection */
        $productCollection = $this->productCollectionFactory->create();
        $productCollection->setStore($store);
        $productCollection->setPageSize($pageSize);
        $productCollection->setCurPage($page);

        $productCollection->addAttributeToSelect(
            array_merge(
                $this->helper->getAllAttributesToCheck(),
                ['price', 'status', 'to_check', 'manual_status']
            )
        );

        foreach ($this->helper->getAttributeCheckAllowedTypes() as $allowedType) {
            if ($this->helper->canCheckImage($allowedType)) {
                $productCollection->addMediaGalleryData();
            }

            if ($this->helper->canCheckCategories($allowedType)) {
                $productCollection->addCategoryIds();
            }
        }

        $productCollection->addFieldToFilter(
            [
                ['attribute' => 'manual_status', 'neq' => 1],
                ['attribute' => 'to_check', 'eq' => 1]
            ]
        );

        return $productCollection;
    }

    /**
     * @param \Magento\Catalog\Model\ResourceModel\Product\Collection $collection
     * @param string $linkField
     * @param \Magento\Store\Api\Data\StoreInterface $store
     * @param \FiloBlu\Checkproducts\Model\CollectorInterface $collector
     * @return void
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function processProducts(
        Collection $collection,
        string $linkField,
        StoreInterface $store,
        CollectorInterface $collector
    ) {
        $toEnable = [];
        $toDisable = [];

        foreach ($collection as $product) {
            /** @var Product $product */
            $valid = $this->validate($product, $collector);

            if ((int)$product->getData('manual_status')) {
                continue;
            }

            if ($valid) {
                $toEnable[] = $product->getData($linkField);
                continue;
            }

            $toDisable[] = $product->getData($linkField);
        }

        if (!empty($toEnable)) {
            $this->updateAttributes(
                $toEnable,
                Status::STATUS_ENABLED,
                $linkField,
                $store->getId()
            );
        }

        if (!empty($toDisable)) {
            $this->updateAttributes(
                $toDisable,
                Status::STATUS_DISABLED,
                $linkField,
                $store->getId()
            );
        }
    }

    /**
     * @param Product|ProductInterface $product
     * @param CollectorInterface $collector
     * @return bool
     * @throws LocalizedException
     */
    public function validate(ProductInterface $product, CollectorInterface $collector): bool
    {
        $type = $product->getTypeId();

        $canCollect = (bool)$product->getData('to_check');
        $valid = true;

        if ($this->helper->canCheckPrice($type) && $this->helper->getPrice($product) <= 0.0) {
            if (!$canCollect) {
                return false;
            }

            $valid = false;
            $collector->collect($product, $type === 'configurable' ? 'price_configurable' : 'price');
        }

        if ($this->helper->canCheckImage($type) && count($product->getMediaGalleryEntries()) === 0) {
            if (!$canCollect) {
                return false;
            }

            $valid = false;
            $collector->collect($product, 'images');
        }

        if ($this->helper->canCheckCategories($type) && count($product->getCategoryIds()) === 0) {
            if (!$canCollect) {
                return false;
            }

            $valid = false;
            $collector->collect($product, 'categories');
        }

        foreach ($this->helper->getAttributesToCheck($type) as $attributeCode) {
            if (!$product->getData($attributeCode)) {
                if (!$canCollect) {
                    return false;
                }

                $valid = false;
                $collector->collect($product, [$attributeCode => null]);
            }
        }

        return $valid;
    }

    /**
     * @param array $toEnable
     * @param $status
     * @param $linkField
     * @param int $storeId
     * @return void
     */
    private function updateAttributes(array $toEnable, $status, $linkField, $storeId)
    {
        $connection = $this->resourceConnection->getConnection();
        $attributeId = $this->eavAttribute->getIdByCode(Product::ENTITY, ProductInterface::STATUS);
        $mainTable = $connection->getTableName('catalog_product_entity_int');

        $data = [];

        foreach ($toEnable as $id) {
            $data[] = [
                $linkField => $id,
                'store_id' => $storeId,
                'value' => $status,
                'attribute_id' => $attributeId
            ];
        }

        if ($connection->getTransactionLevel() !== 0) {
            $connection->insertOnDuplicate($mainTable, $data, ['value']);
            return;
        }

        for ($retries = self::MAX_RETRIES - 1; $retries > 0; $retries--) {
            try {
                $connection->insertOnDuplicate($mainTable, $data, ['value']);
            } /** @noinspection BadExceptionsProcessingInspection */
            catch (DeadlockException $e) {
                usleep(rand(500000, 3500000));
                continue;
            }
        }
    }
}
